__version__ = "6.0.0"
__author__ = "Rob McMullen"
__author_email__ = "feedback@playermissile.com"
__url__ = "https://github.com/robmcmullen/atrcopy"
__bug_report_url__ = "https://github.com/robmcmullen/atrcopy/issues"
